/*
 * Copyright (c) 1998-2006 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 2.0 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 * 
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */

#include <i386/cpuid.h>
#include <IOKit/IOPlatformExpert.h>
#include <IOKit/IOMemoryDescriptor.h>
#include <IOKit/IODeviceTreeSupport.h>
#include "AppleSMBIOSEFI.h"

#if     DEBUG
#define DEBUG_LOG(args...)  IOLog(args)
#else
#define DEBUG_LOG(args...)
#endif

//---------------------------------------------------------------------------
// SMBPackedStrings class

class SMBPackedStrings 
{
protected:
    const char * _start;
    const char * _limit;

public:
    SMBPackedStrings( const SMBStructHeader * header, const void * limit );
	SMBPackedStrings( const SMBStructHeader * header );

    const char * stringAtIndex( UInt8 index, UInt8 * length = 0 ) const;

    void setDataProperty( IORegistryEntry * entry,
                          const char *      key,
                          UInt8             index ) const;

    void setStringProperty( IORegistryEntry * entry,
                            const char *      key,
                            UInt8             index ) const;
};

SMBPackedStrings::SMBPackedStrings( const SMBStructHeader * header,
                                    const void *            limit )
{
    _start = (const char *) header + header->length;
    _limit = (const char *) limit;
}

SMBPackedStrings::SMBPackedStrings( const SMBStructHeader * header )
{
    _start = (const char *) header + header->length;
	
	const char * cp = _start;
	// Find the double null at the end of the record
	while( cp[0] || cp[1]) cp++;
	
	_limit = &cp[1];
}
    
const char * SMBPackedStrings::stringAtIndex( UInt8 index, UInt8 * length ) const
{
    const char * last = 0;
    const char * next = _start;

    if ( length ) *length = 0;

    while ( index-- )
    {
        last = 0;
        for ( const char * cp = next; cp < _limit; cp++ )
        {
            if ( *cp == '\0' )
            {
                last = next;
                next = cp + 1;
                break;
            }
        }
        if ( last == 0 ) break;
    }

    if ( last )
    {
        while (*last == ' ') last++;
        if (length)
        {
            UInt8 len;
            for ( len = next - last - 1; len && last[len - 1] == ' '; len-- )
                ;
            *length = len; // number of chars not counting the terminating NULL
        }
    }

    return last ? last : "";
}

void SMBPackedStrings::setDataProperty( IORegistryEntry * entry,
                                        const char *      key,
                                        UInt8             index ) const
{
    UInt8        length;
    const char * string = SMBPackedStrings::stringAtIndex(index, &length);
    
    if (length)
    {
        OSData * data = OSData::withCapacity(length + 1);
        if (data)
        {
            data->appendBytes(string, length);
            data->appendByte('\0', 1);
            entry->setProperty(key, data);
            data->release();
        }
    }
}

void SMBPackedStrings::setStringProperty( IORegistryEntry * entry,
                                          const char *      key,
                                          UInt8             index ) const
{
    UInt8        length;
    const char * string = SMBPackedStrings::stringAtIndex(index, &length);

    if (length)
    {
        OSString * strObj = OSString::withCString(string);
        if (strObj)
        {
            entry->setProperty(key, strObj);
            strObj->release();
        }
    }
}


#define super IOService

OSDefineMetaClassAndStructors( AppleSMBIOSEFI, IOService )

//---------------------------------------------------------------------------

bool AppleSMBIOSEFI::start( IOService * provider )
{
    if (super::start(provider) != true     ||
        IOService::getResourceService() == 0 ||
        IOService::getResourceService()->getProperty("SMBIOS") == 0)
    {
        return false;
    }
	
	SMBIOSTable = NULL;
	SMBIOSTableLength = 0;

    memInfoSource = kNoMemoryInfo;
    memSlotsData  = OSData::withCapacity(kMemDataSize);
    memTypesData  = OSData::withCapacity(kMemDataSize);
    memSizesData  = OSData::withCapacity(kMemDataSize);
    memSpeedData  = OSData::withCapacity(kMemDataSize);
    memInfoData   = OSData::withCapacity(kMemDataSize);
    memManufData  = OSData::withCapacity(kMemDataSize);
    memSerialData = OSData::withCapacity(kMemDataSize);
    memPartData   = OSData::withCapacity(kMemDataSize);
    memSizeTotal  = 0;

    if (!memSlotsData || !memTypesData || !memSizesData || !memSpeedData ||
        !memInfoData || !memManufData || !memSerialData || !memPartData)
        return false;

	if (!findSMBIOSTableEFI())
	{
		return false;
	}
	else
	{
        IOLog("SMBIOS: EFI bootloader\n");
	}

    // Update device tree
    updateDeviceTree();
	
    return false;
}

//---------------------------------------------------------------------------

#define RELEASE(x) do { if (x) { (x)->release(); (x) = 0; } } while(0)

void AppleSMBIOSEFI::free( void )
{
    RELEASE( memSlotsData  );
    RELEASE( memTypesData  );
    RELEASE( memSizesData  );
    RELEASE( memSpeedData  );
    RELEASE( memInfoData   );
    RELEASE( memManufData  );
    RELEASE( memSerialData );
    RELEASE( memPartData   );
    RELEASE( fDMIMemoryMap );

    super::free();
}

//---------------------------------------------------------------------------

static UInt8 checksum8( void * start, UInt length )
{
    UInt8   csum = 0;
    UInt8 * cp = (UInt8 *) start;

    for (UInt i = 0; i < length; i++)
        csum += *cp++;

    return csum;
}

#define EFI_SMBIOS_TABLE \
"/efi/configuration-table/EB9D2D31-2D88-11D3-9A16-0090273FC14D"

//---------------------------------------------------------------------------

bool AppleSMBIOSEFI::findSMBIOSTableEFI( void )
{
	IORegistryEntry *		tableEntry;
	OSData *				tableData;
	UInt64					tableAddr;
	IOMemoryDescriptor *    epsMemory;
	SMBEntryPoint			eps;
	IOMemoryDescriptor *    dmiMemory = 0;
	IOItemCount             dmiStructureCount = 0;

	tableEntry = fromPath(EFI_SMBIOS_TABLE, gIODTPlane);	
	if (tableEntry)
	{		
		tableAddr = 0;
		tableData = OSDynamicCast(OSData, tableEntry->getProperty("table"));
		if (tableData && (tableData->getLength() <= sizeof(tableAddr)))
		{
			bcopy(tableData->getBytesNoCopy(), &tableAddr, tableData->getLength());

			epsMemory = IOMemoryDescriptor::withAddress(
						(mach_vm_address_t) tableAddr,
						(mach_vm_size_t) sizeof(eps),
						kIODirectionOutIn,
						NULL );

			if (epsMemory)
			{
				bzero(&eps, sizeof(eps));
				epsMemory->readBytes(0, &eps, sizeof(eps));

				if (memcmp(eps.anchor, "_SM_", 4) == 0)
				{
					UInt8 csum;

					csum = checksum8(&eps, sizeof(eps));

					DEBUG_LOG("DMI checksum       = 0x%x\n", csum);
					DEBUG_LOG("DMI tableLength    = %d\n",
						eps.dmi.tableLength);
					DEBUG_LOG("DMI tableAddress   = 0x%lx\n",
						eps.dmi.tableAddress);
					DEBUG_LOG("DMI structureCount = %d\n",
						eps.dmi.structureCount);
					DEBUG_LOG("DMI bcdRevision    = %x\n",
						eps.dmi.bcdRevision);

					if (csum == 0 && eps.dmi.tableLength &&
						eps.dmi.structureCount)
					{
						dmiStructureCount = eps.dmi.structureCount;
						dmiMemory = IOMemoryDescriptor::withPhysicalAddress(
									eps.dmi.tableAddress, eps.dmi.tableLength,
									kIODirectionOutIn );
					}
					else
					{
						DEBUG_LOG("No DMI structure found\n");
					}
				}

				epsMemory->release();
				epsMemory = 0;
			}
		}

		tableEntry->release();
	}

    if ( dmiMemory )
    {
        fDMIMemoryMap = dmiMemory->map();
        if (fDMIMemoryMap)
        {
			SMBIOSTable = (void *) fDMIMemoryMap->getVirtualAddress();
			SMBIOSTableLength =  fDMIMemoryMap->getLength();
			
            decodeSMBIOSTable((void *) fDMIMemoryMap->getVirtualAddress(),
                fDMIMemoryMap->getLength(), dmiStructureCount );
        }
        dmiMemory->release();
        dmiMemory = 0;
    }

    return (fDMIMemoryMap != 0);
}

//---------------------------------------------------------------------------

void AppleSMBIOSEFI::decodeSMBIOSTable( const void * tableData,
                                     UInt16       tableLength,
                                     UInt16       structureCount )
{
    const SMBStructHeader * header;
    const UInt8 *           next = (const UInt8 *) tableData;
    const UInt8 *           end  = next + tableLength;

    while ( structureCount-- && (end > next + sizeof(SMBStructHeader)) )
    {
        header = (const SMBStructHeader *) next;
        if (header->length > end - next) break;

        decodeSMBIOSStructure( header, end );

        // Skip the formatted area of the structure.

        next += header->length;

        // Skip the unformatted structure area at the end (strings).
        // Look for a terminating double NULL.

        for ( ; end > next + sizeof(SMBStructHeader); next++ )
        {
            if ( next[0] == 0 && next[1] == 0 )
            {
                next += 2; break;
            }
        }
    }
}

//---------------------------------------------------------------------------

void AppleSMBIOSEFI::
decodeSMBIOSStructure( const SMBStructHeader * structureHeader,
                       const void *            tableBoundary )
{
    const union SMBStructUnion {
        SMBBIOSInformation      bios;
        SMBSystemInformation    system;
        SMBSystemEnclosure      enclosure;
        SMBProcessorInformation cpu;
        SMBCacheInformation     cache;
        SMBPhysicalMemoryArray  memoryArray;
        SMBMemoryDevice         memoryDevice;
        SMBMemoryModule         memoryModule;
        SMBSystemSlot           slot;
		SMBFirmwareVolume       fv;
		SMBMemorySPD            spd;
        SMBOemProcessorType     oemCpu;
    } * u = (const SMBStructUnion *) structureHeader;

    SMBPackedStrings strings = SMBPackedStrings( structureHeader,
                                                 tableBoundary );

    switch ( structureHeader->type )
    {
        case kSMBTypeBIOSInformation:
            processSMBIOSStructure( &u->bios, &strings );
            break;

        case kSMBTypeSystemInformation:
            processSMBIOSStructure( &u->system, &strings );
            break;

        case kSMBTypeProcessorInformation:
            processSMBIOSStructure( &u->cpu, &strings );
            break;

        case kSMBTypeMemoryDevice:
            processSMBIOSStructure( &u->memoryDevice, &strings );
            break;

        case kSMBTypeMemoryModule:
            processSMBIOSStructure( &u->memoryModule, &strings );
            break;
    }
}

//---------------------------------------------------------------------------
void AppleSMBIOSEFI::
processSMBIOSStructure( const SMBBIOSInformation * bios,
                        SMBPackedStrings * strings )
{
	if (fDMIMemoryMap)
	{
	IORegistryEntry * romNode = fromPath("/rom", gIODTPlane);
	if (romNode)
	{
		OSData * rom_vendor = OSData::withCapacity(10);
		if (rom_vendor)
		{
			rom_vendor->appendBytes("Apple Inc.", 10);
			rom_vendor->appendBytes(NULL, 1);
			romNode->setProperty("vendor", rom_vendor);
			rom_vendor->release();
		}

		OSData * rom_version = OSData::withCapacity(28);
		if (rom_version)
		{
			rom_version->appendBytes("MP31.88Z.006C.B05.0802291410", 28);
			rom_version->appendBytes(NULL, 1);
			romNode->setProperty("version", rom_version);
			rom_version->release();
		}

		romNode->release();
	}
	}
}

//---------------------------------------------------------------------------

void AppleSMBIOSEFI::
processSMBIOSStructure( const SMBSystemInformation * sys,
                        SMBPackedStrings * strings )
{
	if (fDMIMemoryMap)
	{
	IORegistryEntry * rootNode = fromPath("/", gIODTPlane);
	if (rootNode)
	{
		OSData * manufacturer = OSData::withCapacity(10);
		if (manufacturer)
		{
			manufacturer->appendBytes("Apple Inc.", 10);
			manufacturer->appendBytes(NULL, 1);
			rootNode->setProperty("manufacturer", manufacturer);
			manufacturer->release();
		}

		OSData * productName = OSData::withCapacity(9);
		if (productName)
		{
			productName->appendBytes("MacPro3,1", 9);
			productName->appendBytes(NULL, 1);

			const char *productNameString = (char *) productName->getBytesNoCopy();
			
			rootNode->setProperty("product-name", productName);
			rootNode->setProperty("compatible", productName);
			rootNode->setProperty("model", productName);
			rootNode->setName(productNameString);
			productName->release();
		}

		OSData * version = OSData::withCapacity(3);
		if (version)
		{
			version->appendBytes("1.0", 3);
			version->appendBytes(NULL, 1);
			rootNode->setProperty("version", version);
			version->release();
		}
#if 1
		OSData * serialNumber = OSData::withCapacity(11);
		if (serialNumber)
		{
			serialNumber->appendBytes("W1234567890", 11);
			serialNumber->appendBytes(NULL, 1);
			
			UInt8 length = serialNumber->getLength() - 1;
			const char *serialNumberString = (char *) serialNumber->getBytesNoCopy();
			
			OSData * data = OSData::withCapacity(43);
			if (data)
			{
				data->appendBytes(serialNumberString + (length - 3), 3);
				data->appendBytes(NULL, 10);
				data->appendBytes(serialNumberString, length);
				data->appendBytes(NULL, 43 - length - 10 - 3);
				rootNode->setProperty("serial-number", data);
				data->release();
			}
			rootNode->setProperty("IOPlatformSerialNumber", serialNumberString);

			serialNumber->release();
		}

		OSData * boardId = OSData::withCapacity(12);
		if (boardId)
		{
			boardId->appendBytes("Mac-F42C88C8", 12);
			boardId->appendBytes(NULL, 1);
			rootNode->setProperty("board-id", boardId);

			char dummy = '\0';
			rootNode->setProperty("system-type", &dummy, 1);
			boardId->release();
		}
#endif
		rootNode->release();
	}
	}
}

//---------------------------------------------------------------------------

void AppleSMBIOSEFI::
processSMBIOSStructure( const SMBProcessorInformation * cpu,
                        SMBPackedStrings * strings )
{
	UInt64 frequency = 0;

	if (fDMIMemoryMap)
	{
	IORegistryEntry * efiNode = fromPath("/efi/platform", gIODTPlane);
	if(efiNode)
	{
		OSData * FSBFrequency = OSDynamicCast(OSData, efiNode->getProperty("FSBFrequency"));
		if (FSBFrequency)
		{
			frequency = *(UInt64 *) FSBFrequency->getBytesNoCopy();
			frequency += (500ULL * 1000ULL);
			frequency /= (1000ULL * 1000ULL);
		}
		efiNode->release();
	}
	}

    if (cpu->header.length < 26)
        return;

    UInt32 busClockRateMHz = cpu->externalClock > 0 ? cpu->externalClock : (UInt32)frequency;
    UInt32 cpuClockRateMHz = cpu->currentClock;

    if (PE_parse_boot_arg("FSB", &busClockRateMHz)) busClockRateMHz = busClockRateMHz;

    if (busClockRateMHz > 0) {

        i386_cpu_info_t * cpuinfo = cpuid_info();
        IOService *       platform = getPlatform();

        if (!strncmp(cpuinfo->cpuid_vendor, CPUID_VID_INTEL, 
            sizeof(CPUID_VID_INTEL)) &&
            (cpuinfo->cpuid_features & CPUID_FEATURE_SSE2))
        {
            // convert from FSB to quad-pumped bus speed
            busClockRateMHz *= 4;
        }

        platform->callPlatformFunction(
								"SetBusClockRateMHz",     /* function */
								false,                    /* waitForFunction */
								(void *) busClockRateMHz, /* speed in MHz */
								(void *) cpuClockRateMHz,
								0,
								0);
    }
}

//---------------------------------------------------------------------------
#pragma mark Type 17 - kSMBTypeMemoryDevice

void AppleSMBIOSEFI::
processSMBIOSStructure( const SMBMemoryDevice * memory,
                        SMBPackedStrings * strings )
{
    UInt8        deviceLocatorLength;
    const char * deviceLocatorString;
    UInt8        bankLocatorLength;
    const char * bankLocatorString;
    UInt8        memoryType;
    UInt16       memorySpeed;
	char         textString[kMemDataSize];

    union {
        UInt64 ull;
        UInt32 ul[2];
    } memoryBytes;

    if (memory->header.length < 21)
        return;

    if (memInfoSource == kMemoryModuleInfo)
    {
        memSlotsData->initWithCapacity(kMemDataSize);
        memTypesData->initWithCapacity(kMemDataSize);
        memSizesData->initWithCapacity(kMemDataSize);
        memSpeedData->initWithCapacity(kMemDataSize);
        memManufData->initWithCapacity(kMemDataSize);
        memSerialData->initWithCapacity(kMemDataSize);
        memPartData->initWithCapacity(kMemDataSize);
        memSizeTotal = 0;
    }

    memInfoSource = kMemoryDeviceInfo;

    // update memSlotsData

    deviceLocatorString = strings->stringAtIndex( memory->deviceLocator, 
                                                  &deviceLocatorLength );

    bankLocatorString = strings->stringAtIndex( memory->bankLocator, 
                                                &bankLocatorLength );

	if (!memory->deviceLocator)
	{
		deviceLocatorString = "Location";
		deviceLocatorLength = strlen(deviceLocatorString);
	}
	if (!memory->bankLocator)
	{
		bankLocatorString = "Bank";
		bankLocatorLength = strlen(bankLocatorString);
	}

    if ( deviceLocatorLength || bankLocatorString )
    {
        if ( memSlotsData->getLength() == 0 )
		   memSlotsData->appendBytes("   ", 4);
        if ( bankLocatorLength )
            memSlotsData->appendBytes( bankLocatorString, bankLocatorLength );
        if ( deviceLocatorLength && bankLocatorLength )
            memSlotsData->appendByte('/', 1);
        if ( deviceLocatorLength )
            memSlotsData->appendBytes( deviceLocatorString, deviceLocatorLength );
        memSlotsData->appendByte('\0', 1);
    }

    // update memTypesData

#if 0 // disabled, information adds little value
    for ( int bit = 15; bit >= 0; bit-- )
    {
        if ( ( memory->memoryTypeDetail & (1 << bit) ) &&
             ( SMBMemoryTypeDetailStringTable[bit] != NULL ) )
        {
            memTypesData->appendBytes( SMBMemoryTypeDetailStringTable[bit],
                                strlen(SMBMemoryTypeDetailStringTable[bit]));
            memTypesData->appendByte( ' ', 1 );
        }
    }
#endif /* 0 */

	memoryType = memory->memoryType;
	if ( memoryType > kSMBMemoryDeviceTypeCount - 1 )
		memoryType = 0x02; // unknown type

    if (PE_parse_boot_arg("memoryType", &memoryType)) memoryType = memoryType;

	memTypesData->appendBytes( SMBMemoryDeviceTypes[memoryType],
						strlen(SMBMemoryDeviceTypes[memoryType]) + 1 );

    // update memSizesData

    memoryBytes.ull = (memory->memorySize & 0x7fff) * 1024;
    if ((memory->memorySize & 0x8000) == 0)
        memoryBytes.ull *= 1024;

    memSizeTotal += memoryBytes.ull;
    memSizesData->appendBytes( &memoryBytes.ul[1], 4 );
    memSizesData->appendBytes( &memoryBytes.ul[0], 4 );

    // update memSpeedData

	memorySpeed = memory->memorySpeed;
	if ( memorySpeed > 10000 )
		memorySpeed = 0x00; // unknown speed

    if (PE_parse_boot_arg("memorySpeed", &memorySpeed)) memorySpeed = memorySpeed;

	snprintf(textString, sizeof(textString), "%u MHz", memorySpeed);
	memSpeedData->appendBytes(textString, strlen(textString) + 1);

    // update memManufData, memSerialData, memPartData

	snprintf(textString, sizeof(textString), "N/A");
	memManufData->appendBytes(textString, strlen(textString) + 1);

	snprintf(textString, sizeof(textString), "N/A");
	memSerialData->appendBytes(textString, strlen(textString) + 1);

	snprintf(textString, sizeof(textString), "N/A");
	memPartData->appendBytes(textString, strlen(textString) + 1);
}

//---------------------------------------------------------------------------
#pragma mark Type 6 - kSMBTypeMemoryModule

void AppleSMBIOSEFI::
processSMBIOSStructure( const SMBMemoryModule * memory,
                        SMBPackedStrings * strings )
{
    UInt8        socketLength;
    const char * socketString;
    UInt8        memorySize;

    union {
        UInt64 ull;
        UInt32 ul[2];
    } memoryBytes;

    if (memory->header.length < sizeof(SMBMemoryModule))
        return;

    if (memInfoSource == kMemoryDeviceInfo)
        return;

    memInfoSource = kMemoryModuleInfo;

    // update memSlotsData

    socketString = strings->stringAtIndex( memory->socketDesignation, 
                                           &socketLength );

    if ( socketString )
    {
        if (memSlotsData->getLength() == 0)
            memSlotsData->appendBytes("   ", 4);
        if (socketLength)
            memSlotsData->appendBytes( socketString, socketLength );
        memSlotsData->appendByte('\0', 1);
    }

    // update memTypesData

    memTypesData->appendBytes("DRAM", 5);

    // update memSizesData

    memorySize = memory->enabledSize & 0x7F;
    if (memorySize >= kSMBMemoryModuleSizeNotDeterminable)
        memoryBytes.ull = 0;
    else
        memoryBytes.ull = (1ULL << memorySize) * (1024 * 1024);

    memSizeTotal += memoryBytes.ull;
    memSizesData->appendBytes( &memoryBytes.ul[1], 4 );
    memSizesData->appendBytes( &memoryBytes.ul[0], 4 );

    // update memSpeedData

    memSpeedData->appendBytes("0 MHz", 6);
}

//---------------------------------------------------------------------------

void AppleSMBIOSEFI::updateDeviceTree( void )
{
	if (fDMIMemoryMap)
	{
	IORegistryEntry * memoryNode = fromPath("/memory", gIODTPlane);
    if (memoryNode)
    {
        //memoryNode->setName("memory");
        //memoryNode->setLocation("0");
        memoryNode->setProperty( "slot-names",  memSlotsData );
        memoryNode->setProperty( "dimm-types",  memTypesData );
        memoryNode->setProperty( "reg",         memSizesData );
        if (memSpeedData->getLength())
        {
            memoryNode->setProperty( "dimm-speeds", memSpeedData );
		}

		if (memInfoData->getLength() == 0)
		{
            memInfoData->appendBytes(0, (memSizesData->getLength() / 8) * 128);
		}

		memoryNode->setProperty( "dimm-info",          memInfoData );

		if (memManufData->getLength())
			memoryNode->setProperty( "dimm-manufacturer",  memManufData );

		if (memSerialData->getLength())
			memoryNode->setProperty( "dimm-serial-number", memSerialData );

		if (memPartData->getLength())
			memoryNode->setProperty( "dimm-part-number",   memPartData );

        memoryNode->release();
    }
    }
}